/*! ********************************************
 * GrayBox.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_GrayBox_h
#define INCLUDE_GrayBox_h

#ifndef INCLUDE_GrayBox_Types_h
#include "Types.h"
#endif

namespace sora {
namespace graybox {
/*
	[ Channel Message ]
		NoteOff
			type    = 0x80
			channel = 0 - 15
			value0  = 0 - 127   NoteNumber
			value1  = 0
			
		NoteOn
			type    = 0x90
			channel = 0 - 15
			value0  = 0 - 127   NoteNumber
			value1  = 0 - 127   Velocity
			
		Key AfterTouch
			type    = 0xA0
			channel = 0 - 15
			value0  = 0 - 127   NoteNumber
			value1  = 0 - 127   Value
			
		Control Change
			type    = 0xB0
			channel = 0 - 15
			value0  = 0 - 127   CC Number
			value1  = 0 - 127   Value
			
		Program Change
			type    = 0xC0
			channel = 0 - 15
			value0  = 0 - 127   ProgramNumber
			value1  = 0
			
		Channel AfterTouch
			type    = 0xD0
			channel = 0 - 15
			value0  = 0 - 127   Value
			value1  = 0
			
		PitchBend
			type    = 0xE0
			channel = 0 - 15
			value0  = 0 - 127   Value  LSB
			value1  = 0 - 127   Value  MSB
		
	[ System Common Message ]
		System Exclusive (*)
			type    = 0xF0
			channel = 0
			value0  = 0
			value1  = 0
			data    = data... F7[EOX]  (SysEx Message)
			
		System Exclusive (*)
			type    = 0xF7
			channel = 0
			value0  = 0
			value1  = 0
			data    = data... F7[EOX]  (SysEx Message)
			
			(*) Avoid using Multi-Packet SysExMessage, if possible..
				SysEx Event 1 (Packet 1) : F0  data....
				SysEx Event 2 (Packet 2) : F7  data....
				SysEx Event 3 (Packet 3) : F7  data....F7[EOX]
		
		MIDI Time Code 
			type    = 0xF1
			channel = 0
			value0  = 0 - 127
			value1  = 0
		
		Song Position
			type    = 0xF2
			channel = 0
			value0  = 0 - 127   Value  LSB
			value1  = 0 - 127   Value  MSB
		
		Song Select
			type    = 0xF3
			channel = 0
			value0  = 0 - 127   SongNumber
			value1  = 0
		
		Tune Request
			type    = 0xF6
			channel = 0
			value0  = 0
			value1  = 0
		
	[ System RealTime Message ]
			type    = 0xF8 ~ 0xFF
			channel = 0
			value0  = 0
			value1  = 0
		
	[ Meta Message (SMF Track Event) ]
		type    = MetaType [0xFF00 - 0xFF7F]
		channel = 0
		value0  = 0 or value(*)
		value1  = 0
		data    = data... (Meta Event Data)
		
		(*)
			FF00  sequence number [0 - 65535]
			FF20  channel prefix  [0 - 15]
			FF2F  end of track    [0]
			FF51  set tempo       [0 - 16777215]
*/
class MidiMessage
{
public:
	static constexpr int32 DATA_PACKET_SIZE = 256;

	int32 type    = 0x80;  // * May contain NonStandard MidiMessages.
	int32 channel = 0;
	int32 value0  = 0;
	int32 value1  = 0;

	int32 dataSize = 0;
	uint8 data[DATA_PACKET_SIZE] = {0u};  // include EOX "F7"
	
	virtual void clear();
// Constructor
	//MIDIEvent() = default;
};

class MidiEvent : public MidiMessage
{
public:
	int32 midiPorts = 0xFFFF;  // 16ports. 16BitFlags[15,14,13...2,1,0]
	
	MidiEvent* next = nullptr;
	int32 offsetFrames = 0;

	void clear() override;
// Constructor
	//MIDIEvent() = default;
};
/***********************************************
	Class	ProcessData
************************************************/
class ProcessData
{
public:
	float sampleRate = 44100.0f;
	int32 audioChannels = 2;
	int32 blockSize = 2048;
	float** audioOut = nullptr;
	MidiEvent* midiEventList = nullptr;
	int32 offsetFrames = 0;
	int32 sampleFrames = 0;
	float audioLatency = 0.0f;
// Method
	MidiEvent* getMidiEvents();
	void addToMidiEventList(MidiEvent& event, bool check = false);
// Constructor
	//ProcessData();
};
/***********************************************
	Class	IGrayBoxModule
************************************************/
class GrayBox;
class IGrayBoxModule
{
public:
// host ---	
	virtual void updateSampleRate(float sampleRate) = 0;
	virtual void updateAudioChannels(int32 ch) = 0;
	virtual void updateProcessBlockSize(int32 size) = 0;
// module ---	
	virtual void resetModule() = 0;
	virtual void initiModule() = 0;
	virtual void connectModule(GrayBox* host) = 0;
	virtual void disconnectModule() = 0;
	virtual void updateBypassStatus(bool bypass) = 0;
// process ---	
	virtual void startProcess() = 0;
	virtual void stopProcess() = 0;
	virtual void suspendProcess() = 0;
	virtual void resumeProcess() = 0;
	virtual void process(ProcessData* procData) = 0;
	
	virtual ~IGrayBoxModule();
};
/***********************************************
	Class	GrayBox
************************************************/
class GrayBox
{
protected:
	MidiEvent _eventList;
	ProcessData _processData;
	int32 _moduleSlots = 0;
	IGrayBoxModule** _modules = nullptr;
	bool* _bypassFlag = nullptr;
	float _defAudioLatency = 0.0f;
	float _audioLatency = 0.0f;
// Main System ---
public:
	float getSampleRate(){ return _processData.sampleRate; }
	float setSampleRate(float sampleRate);
	int32 getAudioChannels(){ return _processData.audioChannels; }
	int32 setAudioChannels(int32 ch);
	int32 getProcessBlockSize(){ return _processData.blockSize; }
	int32 setProcessBlockSize(int32 size);
	float getDefaultAudioLatency(){ return _defAudioLatency; }
	float setDefaultAudioLatency(float val){ _defAudioLatency = val; }
	void audioLatency(float val){ _audioLatency = val; }
	void initialize();
// Module ---
	int32 getModuleSlots(){ return _moduleSlots; }
	int32 setModuleSlots(int32 num);
	IGrayBoxModule* getModule(int32 index);
	IGrayBoxModule* attachModule(int32 index, IGrayBoxModule* module);
	IGrayBoxModule* detachModule(int32 index);
	bool getBypassStatus(int32 index);
	void bypassProcess(int32 index, bool bypass);
	void resetModules();
	void initiModules();
// Process ---
	void suspendProcess();
	void resumeProcess();
	void startProcess();
	void stopProcess();
	void _clearProcessData();
	void process(float** audioOut, int32 offsetFrames, int32 sampleFrames);
// Constructor
	GrayBox(int32 blockSize, float sampleRate, int32 audioChannels = 2, int32 moduleSlots = 4);
// Copy Constructor
	GrayBox(const GrayBox&) = delete;
	GrayBox& operator = (const GrayBox&) = delete;
// Move Constructor
	GrayBox(GrayBox&&)noexcept;
	GrayBox& operator = (GrayBox&&)noexcept;
// Destructor
	~GrayBox();
};
}}
#endif