/*! ********************************************
 * SynthesizerPCM.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthesizerPCM_h
#define INCLUDE_SynthesizerPCM_h

#ifndef INCLUDE_MultiTimberSynthBase_h
#include "MultiTimberSynthBase.hpp"
#endif

#ifndef INCLUDE_Tone_h
#include "Tone.hpp"
#endif
#ifndef INCLUDE_SampleData_h
#include "SampleData.hpp"
#endif


namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	SynthesizerPCM
************************************************/
class SynthesizerPCM final : public MultiTimberSynthBase
{
public:
	static constexpr int32 PG_PHASE_FP = 18;  // 18bit fixed-point  (precision 1 / 262144)
/***********************************************
	Class	SynthesizerPCM :: VoiceUnitPCM
************************************************/
private:	
	class VoiceUnitPCM final : public VoiceUnitBase
	{
	// Tables
		// PG
	private:	
		static constexpr int32 _PG_PHASE_FP      = SynthesizerPCM::PG_PHASE_FP;
		static constexpr int32 _PG_PHASE_FP_MASK = (1 << _PG_PHASE_FP) - 1;
		// LFO
		static constexpr int32 _LFO_TIMER_INIT_VALUE = SynthData::LFO_TIMER_INIT_VALUE;
		static constexpr int32 _LFO_WAVE_TABLE_MASK  = SynthData::LFO_WAVE_TABLE_SIZE - 1;
		int32* _LFO_WAVE_TABLE = nullptr;
	// Synthesizer Data
	public:
		VoicePart* VOICE_PARTS = nullptr;
		TonePresetPCM* TONE_PRESETS = nullptr;
		SampleData* SAMPLE_DATAS = nullptr;
		SamplerPreset* SAMPLER_PRESETS = nullptr;
	
		int32 pgSamplePow2tIdx = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
		int32 pgKeyPow2tIdx    = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
		int32 egPow2tIdx       = 0;  // EG PhaseStep (Pow2Table Index Offset)
		int32 egPhaseStepSndOff= 0;
		int32 lfoTimerCoef     = 0;  // Coefficient Value - LFO Phase Update Timer
	// VoiceUnit Status
		// Pitch
	private:	
		int32 _ptBend     = 0;   // pitchBend
		int32 _ptBendSens = 0;   // pitchBend Sensitivity
		// LFO
		int32 _lfoPhase     = 0;  // LFO Phase
		int32 _lfoTimer     = 0;  // LFO Phase Update Timer
		int32 _lfoTimerStep = 0;  // Update Timer Step
		int32 _lfoMod       = 0;  // Modulation Depth
		int32 _lfoModSens   = 0;  // Modulation Sensitivity
		int32 _lfoOutPM     = 0;  // output - phase modulation
		int32 _lfoOutAM     = 0;  // output - amplitude modulation
		// Wave
		int32* _waveTable = nullptr;
		// EG
		/*
		- Envelop Generator State -
		[ 5674 3210 ] 8Bits Flag Status (0 = deActive)
		[ ---- RSDA ] "-" = Reserved Bits
			A = Attack  stage
			D = Decay   stage
			S = Sustain stage
			R = Release stage
		*/
		int32 _egState      = 0;
		int32* _egLevelTable= nullptr;
		int32 _egCount      = 0;
		int32 _egCountStep  = 0;
		//int32 _egCountStepAR= 0;
		int32 _egCountStepDR= 0;
		int32 _egCountStepRR= 0;
		int32 _egCntEnvSus  = 0;
		int32 _egCntEnvBtm  = 0;
		int32 _egGate       = 0;
		int32 _egSL         = 0;
		int32 _egTL         = 0;
		int32 _egTLv        = 0;
		// PG
		int32 _pgKey        = 0;
		int32 _pgBend       = 0;
		int32 _pgPhase      = 0;
		int32 _pgPhaseAFP   = 0;  // PG Phase (After the Fixed Point)
		int32 _pgPhaseStep  = 0;
	// Sampler
		int32 _plbkE      = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
		int32 _loopE      = 0;  // Loop End Frame       [loopE <= plbkE]
		int32 _loopF      = 0;  // Loop Frames          [val = loopE - lpStart]
		int32 _loopFk     = 0;  // Loop Frames          * after -> keyOff
	
	public:
		void keyOff() override;
		void soundOff() override;
	private:
		int32 _shiftEGState();
	public:
		void keyOn(const int32&, const int32&, const int32&) override;
		void keyOnSampler(const int32&, const int32&, const int32&);
		void updPartBendSens(const int32&) override;
		void updPartPitchBend(const int32&) override;
		void updPartModulation(const int32&) override;
		void updateLfoFrequency(const int32&);
	// Constructor
	public:
		VoiceUnitPCM();
	// process ---
	private:
		void _proc0(float* const&, float* const&, const int32&) override;
		void _proc1(float* const&, float* const&, const int32&) override;
		void _proc2(float* const&, float* const&, const int32&) override;
		void _proc3(float* const&, float* const&, const int32&) override;
		void _proc4(float* const&, float* const&, const int32&) override;
		void _proc5(float* const&, float* const&, const int32&) override;
		void _proc6(float* const&, float* const&, const int32&) override;
		void _proc7(float* const&, float* const&, const int32&) override;
	};

// Factory -----
private:	
	void _recreateVU(int32 maxPoly) override;
	void _resetKeyParams() override;
// Synthesizer - Data -----
private:
	SampleData* SAMPLE_DATAS = nullptr;
public:
	TonePresetPCM* TONE_PRESETS = nullptr;  // [0 - 127](ProgramChange)
	SamplerPreset* SAMPLER_PRESETS = nullptr;
	
	SampleData* getSampleData(int32 id){ return &SAMPLE_DATAS[id]; }
	void setSampleData(int32 id, const SampleData& sampleData, bool deepCopy = true);
	void clearSampleData(int32 id);
	void factoryDataReset() override;
// Synthesizer - main -----
	static constexpr int32 MODEL_ID = 0x23;
private:
	VoiceUnitPCM* _voiceUnitExt = nullptr;
// Constructor
public:
	SynthesizerPCM() : SynthesizerPCM(32, 440.0f, 1.0f){}
	SynthesizerPCM(int32 maxPoly = 32, float refPitch = 440.0f, float masVol = 1.0f);// Controls ---
// Copy Constructor
	SynthesizerPCM(const SynthesizerPCM&) = delete;
	SynthesizerPCM& operator = (const SynthesizerPCM&) = delete;
// Move Constructor
	SynthesizerPCM(SynthesizerPCM&&)noexcept;
	SynthesizerPCM& operator = (SynthesizerPCM&&)noexcept;
// Destructor
	~SynthesizerPCM();
// MidiEvent ---
private:	
	void _noteOn(const MidiEvent&) override;
	void _readSysExData(const MidiEvent&) override;
};

}}}
#endif