/*! ********************************************
 * smf.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_smf_h
#include "smf.hpp"
#endif

namespace sora {
namespace graybox {
namespace smf {
/***********************************************
	Class	SMFTrackEvent
************************************************/

void SMFTrackEvent::clear()
{
	next = nullptr;
	deltaTime = 0;
		
	channel = 0;
	type = 0;
	value0 = 0;
	value1 = 0;
	dataSize = 0;
	if (data != nullptr)delete[] data;
	data = nullptr;
}
SMFTrackEvent::SMFTrackEvent(int32 deltaT, int32 type, int32 ch, int32 val0, int32 val1)
{
	deltaTime = deltaT;
	channel = ch;
	this->type = type;
	value0 = val0;
	value1 = val1;
	if (type == 0x90 && val1 == 0) type = 0x80; // [ noteOn & 0 velocity ] to [ noteOff ]
}
// Copy Constructor
SMFTrackEvent::SMFTrackEvent(const SMFTrackEvent & x)
{
	next = x.next;
	deltaTime = x.deltaTime;

	type = x.type;
	channel = x.channel;
	value0 = x.value0;
	value1 = x.value1;

	dataSize = x.dataSize;
	if (x.data != nullptr)
	{
		data = new uint8[dataSize];
		uint8* _from = x.data;
		uint8* _to = data;
		for (int i = 0; i < dataSize; ++i)_to[i] = _from[i];
	}
}
SMFTrackEvent & SMFTrackEvent::operator=(const SMFTrackEvent & x)
{
	if (this == &x)return *this;
	next = x.next;
	deltaTime = x.deltaTime;

	type = x.type;
	channel = x.channel;
	value0 = x.value0;
	value1 = x.value1;

	dataSize = x.dataSize;
	if (data != nullptr)delete[] data; data = nullptr;
	if (x.data != nullptr)
	{
		data = new uint8[dataSize];
		uint8* _from = x.data;
		uint8* _to = data;
		for (int i = 0; i < dataSize; ++i)_to[i] = _from[i];
	}
	return *this;
}
// Move Constructor
SMFTrackEvent::SMFTrackEvent(SMFTrackEvent && x) noexcept
{
	next = x.next;
	deltaTime = x.deltaTime;

	type = x.type;
	channel = x.channel;
	value0 = x.value0;
	value1 = x.value1;

	dataSize = x.dataSize;
	data = x.data;
	x.data = nullptr;
}
SMFTrackEvent & SMFTrackEvent::operator=(SMFTrackEvent && x) noexcept
{
	if (this == &x)return *this;
	next = x.next;
	deltaTime = x.deltaTime;

	type = x.type;
	channel = x.channel;
	value0 = x.value0;
	value1 = x.value1;

	dataSize = x.dataSize;
	if (data != nullptr)delete[] data;
	data = x.data; x.data = nullptr;
	return *this;
}
SMFTrackEvent::~SMFTrackEvent()
{
	if (data != nullptr)delete[] data;
	data = nullptr;
}
/***********************************************
	Class	SMFSequence
************************************************/
void SMFSequence::clearData()
{
	if (track != nullptr)
	{
		SMFTrackEvent* _track = track;
		SMFTrackEvent* _deleteEv = nullptr;
		while (_track->next != nullptr)
		{
			_deleteEv = _track;
			_track = _track->next;
			delete _deleteEv;
		}
		delete _track; // LastEvent
	}
	track = nullptr;
	division = 480;
	loopable = false;
	loopStart = nullptr;
	loopEnd = nullptr;
}
SMFSequence::SMFSequence(uint8* dat, uint32 dataSize, bool multiPacketSysEx)
{
	loadBytes(dat, dataSize, multiPacketSysEx);
}
// Move Constructor
SMFSequence::SMFSequence(SMFSequence && x) noexcept
{
	_format = x._format;
	track = x.track; x.track = nullptr;
	division = x.division;
	loopable = x.loopable;
	loopStart = x.loopStart;
	loopEnd = x.loopEnd;
}
SMFSequence & SMFSequence::operator=(SMFSequence && x) noexcept
{
	if (this == &x)return *this;
	clearData();
	_format = x._format;
	track = x.track; x.track = nullptr;
	division = x.division;
	loopable = x.loopable;
	loopStart = x.loopStart;
	loopEnd = x.loopEnd;
	return *this;
}
SMFSequence::~SMFSequence()
{
	if (track != nullptr)
	{
		SMFTrackEvent* _track = track;
		SMFTrackEvent* _deleteEv = nullptr;
		while (_track->next != nullptr)
		{
			_deleteEv = _track;
			_track = _track->next;
			delete _deleteEv;
		}
		delete _track; // LastEvent
	}
	track = nullptr;
}
void SMFSequence::_setLoopPoint()
{
	if (track == nullptr) return;
	loopable = false;
	// serch LoopPoint -->
	SMFTrackEvent* _trackPos = track;
	SMFTrackEvent* _start = nullptr;
	SMFTrackEvent* _end = nullptr;
	bool _first = true;
	while (_trackPos != nullptr)
	{
		if (_trackPos->type == 0xff06 && _trackPos->dataSize >= 4)
		{
			uint8* _data = _trackPos->data;
			uint32 _code = _data[0] << 24u | _data[1] << 16u | _data[2] << 8u | _data[3];
			if (_code == 0x6c6f6f70u) // loop
			{
				if (_first)_start = _trackPos;
				else _end = _trackPos;
				_first = false;
			}
		}
		_trackPos = _trackPos->next;
	}
	// 0 DelayTime?
	if (_start != nullptr && _end != nullptr)
	{
		int _time = 0;
		_trackPos = _start;
		while (_trackPos != _end)
		{
			_time = _trackPos->next->deltaTime;
			if (_time) break;
			_trackPos = _trackPos->next;
		}
		if (_time == 0){ _start = nullptr; _end = nullptr; }
	} else { _start = nullptr; _end = nullptr; }
	// set LoopPoint -->
	loopStart = _start;
	loopEnd = _end;
	loopable = true;
}
bool SMFSequence::loadBytes(uint8* bytes, uint32 dataSize, bool multiPacketSysEx)
{
	// 'MThd' ---
	if (dataSize <= 16u) return false;
	uint32 p = 0u;
	_multiPacketSysEx = multiPacketSysEx;
	uint32 _tag = uint32( bytes[p]<<24u | bytes[p+1u]<<16u | bytes[p+2u]<<8u | bytes[p+3u] ); p+=4u;
	if (_tag != 0x4d546864u) return false;
	p += 4u;
	_format	= uint32( bytes[p]<<8u | bytes[p+1u] ); p+=2u;
	p += 2u; //uint32 _trks = uint32( bytes[p]<<8 | bytes[p+1] ); p+=2;
	int32 _div  = int32( bytes[p]<<8u | bytes[p+1u] ); p+=2u;
	if (_format >= 2 || _div & 0x8000) return false;  // No Suport "Format2", "SMPTE"
	// 'MTrk' ---
	uint32 _chunkType, _chunkDataSize;
	SMFTrackEvent _headEv(0, 0x80, 0, 0);
	SMFTrackEvent* _head = &_headEv;

	while (p < dataSize)
	{
		_chunkType     = uint32( bytes[p]<<24u | bytes[p+1u]<<16u | bytes[p+2u]<<8u | bytes[p+3u] ); p+=4u;
		_chunkDataSize = uint32( bytes[p]<<24u | bytes[p+1u]<<16u | bytes[p+2u]<<8u | bytes[p+3u] ); p+=4u;
		if (_chunkType == 0x4d54726bu)
		{
			if (!_read_MTrk(bytes + p, _chunkDataSize, _head))
			{
				track = _head->next;
				clearData();
				return false;
			}
		}
		// else // Unknown Chunk
		p += _chunkDataSize;
	}
// SetUp -->
	clearData();
	track = _head->next;
	if (_format != 0)_removeUnecEvents(track);
	division = _div;
	_setLoopPoint();
// <-- Finish
	return true;
}
bool SMFSequence::_read_MTrk(uint8* bytes, uint32 chunkDataSize, SMFTrackEvent* trk)
{
	uint8* _cunkEnd = bytes + chunkDataSize;
	
	int32 _eventType = 0, _channel = 0;
	int32 _value0 = 0, _value1 = 0;
	int32 _skipTime = 0;
	SMFTrackEvent* _event = nullptr;

	uint8* _sysExData = nullptr;
	int32 _sysExDataSize;
	uint8* _metaData = nullptr;
	int32 _metaType, _metaDataSize;
	// Start -->


	while (bytes < _cunkEnd)
	{
	// Read DeltaTime ---
		int32 _deltaTime = _readVarLeng(bytes);
		if (_skipTime > 0){ _deltaTime += _skipTime; _skipTime = 0; }
	// Read StatusByte ---
		int32 _val = *(bytes++);
		if (_val & 0x80)
		{
			int32 _status = _val;
			if (_status < 0xF0) { _eventType = _status & 0xf0; _channel = _status & 0xf; }  // Channel Message
			else { _eventType = _status; _channel = 0; }  // SysEx, MetaEvent
		}
		else bytes--;  // Runnig Status
			
		switch (_eventType)
		{
			default :  return false; // Unknown Event
		// MIDI Channel Message ---
			case 0x80 : case 0x90 : case 0xA0 : case 0xB0 : case 0xE0 :
				_value0 = *(bytes++);
				_value1 = *(bytes++);
				_event = new SMFTrackEvent(_deltaTime, _eventType, _channel, _value0, _value1);
				_insertTrackEvent(trk, _event);
				continue;

			case 0xC0 : case 0xD0 :
				_value0 = *(bytes++);
				_event = new SMFTrackEvent(_deltaTime, _eventType, _channel, _value0);
				_insertTrackEvent(trk, _event);
				continue;

		// System Exclusive ---
			case 0xF0 : 
				_event = new SMFTrackEvent(_deltaTime, 0xF0, 0);
				_insertTrackEvent(trk, _event);
				// Multi Packet Message -->
				if (_multiPacketSysEx)
				{
					// Read SysExData
					_sysExDataSize = _readVarLeng(bytes);
					_sysExData = _event->data = new uint8[_sysExDataSize];
					for (int i = 0; i < _sysExDataSize; ++i, ++bytes)_sysExData[i] = *bytes;
					_event->dataSize = _sysExDataSize;
				}
				// Single Packet Message -->
				else
				{
					uint8* _sysExStart = bytes;
					_skipTime = 0;
					// calc DataSize -->
					int _sysExPacketsSize = 0,
						_sysExPackets = 0, 
						_sysExTotalSize = 0;
					while(true)
					{
						_sysExPacketsSize = _readVarLeng(bytes);
						_sysExTotalSize += _sysExPacketsSize;
						bytes += _sysExPacketsSize;  // n Byte seaq
						++_sysExPackets;
						if ((*(bytes - 1) & 0xffu) == 0xF7u) break;  // EOX
						_skipTime += _readVarLeng(bytes);
						++bytes; // status F7
					}
					//uint8* _sysExEnd = bytes;
					// read Data -->
					_sysExData = _event->data = new uint8[_sysExTotalSize];
					_event->dataSize = _sysExTotalSize;
					bytes = _sysExStart;
					while (true)
					{
						_sysExPacketsSize = _readVarLeng(bytes);
						while (_sysExPacketsSize-- > 0)*(_sysExData++) = *(bytes++);
						if (--_sysExPackets > 0) break;
						while (*(bytes++) & 0x80u);  // skip VarLengData
						++bytes; // status F7
					}
					//bytes = _sysExEnd;
				}
				continue;

			case 0xF7 : 
				if (_multiPacketSysEx)
				{
					_event = new SMFTrackEvent(_deltaTime, 0xF7, 0);
					_insertTrackEvent(trk, _event);
					_sysExDataSize = _readVarLeng(bytes);
					_sysExData = _event->data = new uint8[_sysExDataSize];
					for (int i = 0; i < _sysExDataSize; ++i, ++bytes)_sysExData[i] = *bytes;
					_event->dataSize = _sysExDataSize;
				}
				// Non-standard error
				else return false;
				continue;

		// Meta Event ---
			case 0xFF :
				_metaType = *(bytes++) & 0xff;
				_event = new SMFTrackEvent(_deltaTime, 0xFF00 | _metaType, 0);
				_insertTrackEvent(trk, _event);
				// Read MetaData
				_metaDataSize = _readVarLeng(bytes);
				_event->dataSize = _metaDataSize;
				if (_metaDataSize > 0)
				{
					_metaData = _event->data = new uint8[_metaDataSize];
					for (int i = 0; i < _metaDataSize; ++i, ++bytes)_metaData[i] = *bytes;
				}
				switch (_metaType)
				{
					case 0x2F: return true; // End of Track
					case 0x00: _event->value0 = int32(_event->data[0]<<8 | _event->data[1]); break; // Sequence Number
					case 0x20: _event->value0 = int32(_event->data[0]); break; // Channel Prefix
					case 0x51: _event->value0 = int32(_event->data[0]<<16 | _event->data[1]<<8 | _event->data[2]); break; // Set Tempo
				}
				continue;
		}
	}
	// <-- Finish
	return true;
}
void SMFSequence::_insertTrackEvent(SMFTrackEvent* &crnt, SMFTrackEvent* ev)
{
	if (_format == 0)
	{
		crnt->next = ev;
		crnt = ev;
		return;
	}

	SMFTrackEvent* _next = crnt->next;
	while (_next != nullptr)
	{
		if (ev->deltaTime >= _next->deltaTime) ev->deltaTime -= _next->deltaTime;
		else
		{
			_next->deltaTime -= ev->deltaTime;
			ev->next = _next;
			crnt->next = ev;
			crnt = ev;
			return;
		}
		crnt = _next;
		_next = _next->next;
	}
	crnt->next = ev;
	crnt = ev;
}
void SMFSequence::_removeUnecEvents(SMFTrackEvent* trk)
{
	if (trk == nullptr)return;
	SMFTrackEvent* _eotEv = nullptr;
	while (trk->next != nullptr)
	{
		if (trk->next->type == 0xFF2F)
		{
			_eotEv = trk->next;
			if (_eotEv->next == nullptr) return;  // last EOT
			_eotEv->next->deltaTime += _eotEv->deltaTime;
			trk->next = _eotEv->next;
			delete _eotEv;
		}
		trk = trk->next;
	}
}
int32 SMFSequence::_readVarLeng(uint8* &dat)
{
	int32 _v = 0;
	do _v = _v << 7 | *(dat) & 0x7f; while (*(dat++) & 0x80u);
	return _v;
}

/***********************************************
	Class	SMFSequencer
************************************************/
// EventBuffer -----
void SMFSequencer::clearMIDIEventBuffer()
{
	for (int i = 0; i < _eventBufferSize; ++i)_eventBuffer[i].clear();
	_usedEvents = 0;
}
int32 SMFSequencer::setEventBufferSize(int32 size)
{
	if (size == _eventBufferSize) return _eventBufferSize;
	if (size < 64)size = 64;
	delete[] _eventBuffer;
	_eventBuffer = new MidiEvent[_eventBufferSize];
	_usedEvents = 0;
	return _eventBufferSize = size;
}
// Sequence -----
bool SMFSequencer::setSequence(SMFSequence* seq)
{
	reset();
	if (seq->track == nullptr) return false;
	// ---
	_sequence = seq;
	_track = seq->track;
	_division = seq->division;
	_loopable = seq->loopable;
	_loopStart = seq->loopStart;
	_loopEnd = seq->loopEnd;
	// ---
	_refTime = _division * 1000000.0 / _sampleRate;
	_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
	gotoHead();
	return true;
}
// Sequencer status ---
void SMFSequencer::setSampleRate(float val)
{
	_sampleRate = val;
	_refTime = double(_division * 1000000) / _sampleRate;
	_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
}
// Sequencer Contorols ---
void SMFSequencer::reset()
{
		// seq ---
		_tempo = 500000.0;
		_isPlay = false;
		_loopPlay = false;
		_trackPos = nullptr;
		_refTime = 0.0;
		_evTimer = 0;
		_evTimerAFP = 0;
		_evTimerStep = 0;
		gotoHead();
}
void SMFSequencer::initialize()
{
		// data ---
		_sequence = nullptr;
		_track = nullptr;
		_division = 480;
		_loopable = false;
		_loopStart = nullptr;
		_loopEnd = nullptr;
		// seq ---
		_tempo = 500000.0;
		_isPlay = false;
		_loopPlay = false;
		_trackPos = nullptr;
		_refTime = 0.0;
		_evTimer = 0;
		_evTimerAFP = 0;
		_evTimerStep = 0;
}
void SMFSequencer::setBPM(double val)
{
	_tempo = 60000000.0 / val;
	_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
}
// Sequencer Contorols ---
bool SMFSequencer::play(bool resume)
{
	if (_track == nullptr) return false;
	if (!resume || _trackPos == nullptr)gotoHead();
	return _isPlay = true;
}
void SMFSequencer::stop(bool pause)
{
	_isPlay = false;
	if (_track == nullptr) return;
	if (!pause)gotoHead();
}
bool SMFSequencer::loopPlay(bool flag)
{
	if (_loopable) return _loopPlay = flag;
	else return false;
}
void SMFSequencer::gotoHead()
{
	if (_track == nullptr) return;
	_trackPos = _track;
	// ---
	_tempo = 500000.0;
	_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
	// ---
	_evTimer = _trackPos->deltaTime;
	_evTimerAFP = 0;
}
SMFSequencer::SMFSequencer(int32 eventBufferSize)
{
	_eventBufferSize = eventBufferSize;
	_eventBuffer = new MidiEvent[_eventBufferSize];
	_usedEvents = 0;
}
SMFSequencer::SMFSequencer(SMFSequencer && x) noexcept
{
	// EventBuffer -----
	_eventBuffer = x._eventBuffer; x._eventBuffer = nullptr;
	_eventBufferSize = x._eventBufferSize;
	_usedEvents = x._usedEvents;
	// Sequence -----
	_sequence = x._sequence;
	_track = x._track;
	_division = x._division;
	_loopable = x._loopable;
	_loopStart = x._loopStart;
	_loopEnd = x._loopEnd;
	// Sequencer -----
	midiOutPorts = x.midiOutPorts;
	_eventFilter = x._eventFilter;
	_sampleRate = x._sampleRate;
	_tempo = x._tempo;
	_isPlay = x._isPlay;
	_loopPlay = x._loopPlay;
	_trackPos = x._trackPos;
	_refTime = x._refTime;
	_evTimer =  x._evTimer;
	_evTimerAFP = x._evTimerAFP;
	_evTimerStep = x._evTimerStep;
}
SMFSequencer & SMFSequencer::operator=(SMFSequencer && x) noexcept
{
	if (this == &x)return *this;
	// EventBuffer -----
	if (_eventBuffer != nullptr)delete[] _eventBuffer;
	_eventBuffer = x._eventBuffer; x._eventBuffer = nullptr;
	_eventBufferSize = x._eventBufferSize;
	_usedEvents = x._usedEvents;
	// Sequence -----
	_sequence = x._sequence;
	_track = x._track;
	_division = x._division;
	_loopable = x._loopable;
	_loopStart = x._loopStart;
	_loopEnd = x._loopEnd;
	// Sequencer -----
	midiOutPorts = x.midiOutPorts;
	_eventFilter = x._eventFilter;
	_sampleRate = x._sampleRate;
	_tempo = x._tempo;
	_isPlay = x._isPlay;
	_loopPlay = x._loopPlay;
	_trackPos = x._trackPos;
	_refTime = x._refTime;
	_evTimer =  x._evTimer;
	_evTimerAFP = x._evTimerAFP;
	_evTimerStep = x._evTimerStep;
	return *this;
}
SMFSequencer::~SMFSequencer()
{
	if (_eventBuffer != nullptr)delete[] _eventBuffer;
	_eventBuffer = nullptr;
}
// Interface - IfGrayBoxModule ---
void SMFSequencer::updateSampleRate(float sampleRate){ setSampleRate(sampleRate); }

void SMFSequencer::resetModule(){ clearMIDIEventBuffer(); reset(); }
void SMFSequencer::initiModule(){ clearMIDIEventBuffer(); initialize(); }
void SMFSequencer::connectModule(GrayBox* host)
{
	clearMIDIEventBuffer();
	setSampleRate(host->getSampleRate());
}
void SMFSequencer::disconnectModule(){ clearMIDIEventBuffer(); }
void SMFSequencer::updateBypassStatus(bool){ clearMIDIEventBuffer(); }

void SMFSequencer::startProcess(){ clearMIDIEventBuffer(); }
void SMFSequencer::stopProcess(){ clearMIDIEventBuffer(); }
void SMFSequencer::suspendProcess(){ clearMIDIEventBuffer(); }
void SMFSequencer::resumeProcess(){ clearMIDIEventBuffer(); }
void SMFSequencer::process(ProcessData * procData)
{
	if (!_isPlay) return;
	_usedEvents = 0;
	MidiEvent* _list = procData->midiEventList;
	int _sampleFrames = procData->sampleFrames;
	for (int i = 0; i < _sampleFrames; ++i)
	{
		while (_evTimer <= 0)
		{
			// Arbitrary Loop -->
				if (_trackPos == _loopEnd)
				{
					if (_loopPlay)
					{
						_trackPos = _loopStart;
						//if (callBackLoop != null) callBackLoop();
						//continue;
					}
				}
		// Process Event -->
			if (_trackPos->type <= _eventFilter)
			{
				if (_usedEvents < _eventBufferSize)
				{
					MidiEvent& _ev = _eventBuffer[_usedEvents++];
					_ev.next = nullptr;  // clear
					_ev.midiPorts = midiOutPorts;
					_ev.offsetFrames = i;
					_ev.type = _trackPos->type;
					_ev.channel = _trackPos->channel;
					_ev.value0 = _trackPos->value0;
					_ev.value1 = _trackPos->value1;
					if (_trackPos->dataSize > 0)
					{
						int _size = (_trackPos->dataSize > MidiEvent::DATA_PACKET_SIZE) ? 
						    MidiEvent::DATA_PACKET_SIZE : _trackPos->dataSize;
						uint8* _form = _trackPos->data;
						uint8* _to = _ev.data;
						for (int j = 0; j < _size; ++j)_to[j] = _form[j];
						_ev.dataSize = _size;
					}
					else _ev.dataSize = 0;
				}
			}
		// Tempo -->
			if (_trackPos->type == 0xFF51)
			{
				_tempo = _trackPos->value0;
				_evTimerStep = int(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
			}
		// NextEvent, EOT -->
			_trackPos = _trackPos->next;
			if (_trackPos)_evTimer += _trackPos->deltaTime;
			else 
			{
				if (_loopPlay) { gotoHead(); continue; }
				else { _isPlay = false; _evTimer = 1; i = _sampleFrames; }
			}
		}
		// Counter -----
		_evTimer -= _evTimerAFP >> _EV_TIMER_FPBITS;
		_evTimerAFP &= _EV_TIMER_FPMASK;
		_evTimerAFP += _evTimerStep;
	}
	// addtoList -->
	for (int i = 0; i < _usedEvents; )
	{
		MidiEvent& _ev = _eventBuffer[i++];
		while (_list->next)
		{
			if (_ev.offsetFrames >= _list->next->offsetFrames) _list = _list->next;
			else break;
		}
		_ev.next = _list->next;
		_list->next = &_ev;
	}
}
}}}